#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>

#define ATTENTEMAX 10
#define NAME_LENGTH 100
#define MS 1000

void *ecrivain(void *ptr); /* fonction initiale d'un thread écrivain */
void *lecteur(void *ptr); /* fonction initiale d'un thread lecteur */
int donnee = 0; /* ressource commune à tout le monde */
int ecrivains = 0;
int lecteurs = 0;
int valmax = 0;
pthread_mutex_t verrou; /* verrou sur la donnée */
pthread_cond_t condition; /* condition sur la donnée */
int lectures = 0; /* nombre de lectures en cours */
int attente_ecritures = 0; /* nombre d'écritures en attente */
int ecritures = 0; /* nombre d'écritures en cours */

sem_t fin_ecrivains;
sem_t fin_lecteurs;


int main(int argc, char **argv) {
	/* déclaration des variables */
	int i;
	pthread_t ptr_sur_thread;
	char *name;
	/* récupération des arguments */
	if(argc < 4) {
		printf("Usage : %s #lecteurs #ecrivains val_max \n", argv[0]);
		return 0;
	}
	lecteurs = atoi(argv[1]);
	ecrivains = atoi(argv[2]);
	valmax = atoi(argv[3]);


	/* initialisation des variables de synchronisation */
	sem_init(& fin_ecrivains, 0, 0);
	sem_init(& fin_lecteurs, 0, 0);
	pthread_mutex_init(& verrou, NULL);
	pthread_cond_init(& condition, NULL);
	/* initialisation du générateur aléatoire */
	srand(time(NULL)); 

	/* Création de tous les threads */
	for(i = 0; i < lecteurs; i++) {
		name = malloc(sizeof(char) * NAME_LENGTH);
		snprintf(name, NAME_LENGTH, "Lecteur %d", i + 1);
		pthread_create((void *) & ptr_sur_thread, NULL, lecteur, name);
	}
	for(i = 0; i < ecrivains; i++) {
		name = malloc(sizeof(char) * NAME_LENGTH);
		snprintf(name, NAME_LENGTH, "Ecrivain %d", i + 1);
		pthread_create((void *) & ptr_sur_thread, NULL, ecrivain, name);
	}
	/* On attend que tous les lecteurs et que tous les écrivains aient fini */
	for(i = 0; i < lecteurs; i++) sem_wait(& fin_lecteurs);
	for(i = 0; i < ecrivains; i++) sem_wait(& fin_ecrivains);
	return 0;
}

void *ecrivain(void *ptr) {
	char *name = (char*) ptr;
	int attente;
	int donneebis;

	while(donnee < valmax) {
		attente = (int) ((double) rand() * (ATTENTEMAX) / RAND_MAX) + 1;
		usleep(attente * MS);

		printf("%s:\t J'ai envie d'écrire.\n", name);
		/* on entre en section critique et on va boucler tant qu'il y a une lecture en cours ou une écriture en cours */
		pthread_mutex_lock(& verrou);
		attente_ecritures++;
		while((lectures > 0) || (ecritures > 0)) {
			pthread_cond_wait(& condition, & verrou);
		}
		attente_ecritures--;
		ecritures = 1;
		pthread_cond_broadcast(& condition);
		pthread_mutex_unlock(& verrou);

		attente = (int) ((double) rand() * (ATTENTEMAX) / RAND_MAX) + 1;
		printf("%s:\t À mon tour d'écrire (donnee = %d) ! \n", name, donnee);
		donneebis = donnee;
		usleep(attente * MS);
		donneebis++;
		printf("%s:\t Je rends la main (donnee = %d => %d) \n", name, donnee, donneebis);
		donnee = donneebis;
		pthread_mutex_lock(& verrou);
		ecritures = 0;
		pthread_cond_broadcast(& condition);
		pthread_mutex_unlock(& verrou);
	}
	printf("%s :\t Au revoir !\n", name);
	free(name);
	sem_post(& fin_ecrivains);
	return NULL;
}

void *lecteur(void *ptr) {
	char *name = (char*) ptr;
	int attente;
	while(donnee < valmax) {
		attente = (int) ((double) rand() * (3 * ATTENTEMAX) / RAND_MAX) + 1;
		usleep(attente * MS);

		printf("%s :\t J'ai envie de lire !\n", name);
		pthread_mutex_lock(& verrou);
		/* on regarde si on peut prendre place */
		while((attente_ecritures > 0) || (ecritures > 0)) {
			pthread_cond_wait(& condition, & verrou);
		}
		lectures++;
		pthread_cond_broadcast(& condition);
		pthread_mutex_unlock(& verrou);
		printf("%s:\t Je commence à lire... \n", name);
		attente = (int) ((double) rand() * (ATTENTEMAX) / RAND_MAX) + 1;
		usleep(attente * MS);
		printf("%s:\t J'ai pu lire la donnée : %d \n", name, donnee);

		pthread_mutex_lock(& verrou);
		lectures--;
		pthread_cond_broadcast(& condition);
		pthread_mutex_unlock(& verrou);

	}
	printf("%s :\t Au revoir !\n", name);
	free(name);
	sem_post(& fin_lecteurs);
	return NULL;
}

